@@ -1,7 +1,12 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from decimal import Decimal
+
+from sql import operators, Literal
+
+from trytond.i18n import gettext
 from trytond.model import Workflow, ModelView, fields, Check
+from trytond.model.exceptions import AccessError
 from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
 
@@ -10,7 +15,13 @@ __all__ = ['Move']
 
 class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
-    fifo_quantity = fields.Float('FIFO Quantity')
+    fifo_quantity = fields.Float(
+        'FIFO Quantity',
+        help="Quantity used by FIFO.")
+    fifo_quantity_available = fields.Function(fields.Float(
+            "FIFO Quantity Available",
+            help="Quantity available for FIFO"),
+        'get_fifo_quantity_available')
 
     @classmethod
     def __setup__(cls):
@@ -21,17 +32,34 @@ class Move(metaclass=PoolMeta):
         cls._sql_constraints += [
             ('check_fifo_quantity_out',
                 Check(t, t.quantity >= t.fifo_quantity),
-                'FIFO quantity can not be greater than quantity.'),
+                'product_cost_fifo.msg_move_fifo_quantity_greater'),
             ]
-        cls._error_messages.update({
-                'del_move_fifo': ('You can not delete move "%s" that is used '
-                    'for FIFO cost price.'),
-                })
 
     @staticmethod
     def default_fifo_quantity():
         return 0.0
 
+    def get_fifo_quantity_available(self, name):
+        return self.quantity - (self.fifo_quantity or 0)
+
+    @classmethod
+    def domain_fifo_quantity_available(cls, domain, tables):
+        table, _ = tables[None]
+        name, operator, value = domain
+        field = cls.fifo_quantity_available._field
+        Operator = fields.SQL_OPERATORS[operator]
+        column = (
+            cls.quantity.sql_column(table)
+            - cls.fifo_quantity.sql_column(table))
+        expression = Operator(column, field._domain_value(operator, value))
+        if isinstance(expression, operators.In) and not expression.right:
+            expression = Literal(False)
+        elif isinstance(expression, operators.NotIn) and not expression.right:
+            expression = Literal(True)
+        expression = field._domain_add_null(
+            column, operator, value, expression)
+        return expression
+
     def _update_fifo_out_product_cost_price(self):
         '''
         Update the product cost price of the given product on the move. Update
@@ -45,30 +73,26 @@ class Move(metaclass=PoolMeta):
         total_qty = Uom.compute_qty(self.uom, self.quantity,
             self.product.default_uom, round=False)
 
-        fifo_moves = self.product.get_fifo_move(total_qty)
+        with Transaction().set_context(company=self.company.id):
+            fifo_moves = self.product.get_fifo_move(total_qty)
 
         cost_price = Decimal("0.0")
         consumed_qty = 0.0
         to_save = []
         for move, move_qty in fifo_moves:
             consumed_qty += move_qty
-            if move.from_location.type in {'supplier', 'production'}:
-                with Transaction().set_context(date=move.effective_date):
-                    move_unit_price = Currency.compute(
-                        move.currency, move.unit_price,
-                        self.company.currency, round=False)
-                move_unit_price = Uom.compute_price(
-                    move.uom, move_unit_price, move.product.default_uom)
-            else:
-                move_unit_price = move.cost_price or 0
+
+            with Transaction().set_context(date=move.effective_date):
+                move_unit_price = Currency.compute(
+                    move.currency, move.unit_price,
+                    self.company.currency, round=False)
+            move_unit_price = Uom.compute_price(move.uom, move_unit_price,
+                    move.product.default_uom)
             cost_price += move_unit_price * Decimal(str(move_qty))
 
             move_qty = Uom.compute_qty(self.product.default_uom, move_qty,
                     move.uom, round=False)
             move.fifo_quantity = (move.fifo_quantity or 0.0) + move_qty
-            # Due to float, the fifo quantity result can exceed the quantity.
-            assert move.quantity >= move.fifo_quantity - move.uom.rounding
-            move.fifo_quantity = min(move.fifo_quantity, move.quantity)
             to_save.append(move)
         if to_save:
             # TODO save in do method when product change
@@ -118,7 +142,7 @@ class Move(metaclass=PoolMeta):
     @ModelView.button
     @Workflow.transition('cancel')
     def cancel(cls, moves):
-        super().cancel(moves)
+        pass
 
     @classmethod
     def delete(cls, moves):
@@ -127,14 +151,7 @@ class Move(metaclass=PoolMeta):
                 ('fifo_quantity', '!=', 0.0),
                 ])
         if fifo_moves:
-            cls.raise_user_error('del_move_fifo', (fifo_moves[0].rec_name,))
+            raise AccessError(
+                gettext('product_cost_fifo.msg_move_delete_fifo',
+                    move=fifo_moves[0].rec_name))
         super(Move, cls).delete(moves)
-
-    @classmethod
-    def copy(cls, moves, default=None):
-        if default is None:
-            default = {}
-        else:
-            default = default.copy()
-        default.setdefault('fifo_quantity', cls.default_fifo_quantity())
-        return super().copy(moves, default=default)
